/* Declarations for lospa.cpp.
   Copyright (C) 2002,03 Yamamoto Takanori / Tatyana.

This is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <direct.h>

size_t strlcpy(char *dst, const char *src, size_t siz);
size_t strlcat(char *dst, const char *src, size_t siz);
void PrepareDirTree(const char *s0);
size_t Decompress(const unsigned char *src, size_t srclen, unsigned char *dst, size_t dstlen);
void Extract(const char *fn, const union tHeader0 *idx0, const union tHeader1 *idx);

size_t strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

size_t strlcat(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}

union tHeader0{
	unsigned char img[12];
	struct {
		unsigned int idx;	/* number of indexes */
		unsigned int posIndex;	/* abs address of indexes */
		unsigned int posData;	/* abs address of data */
	};
};

union tHeader1{
	unsigned char img[100];
	struct {
		char filename[64];
		unsigned int len;	/* saved data size */
		unsigned int pos;
		unsigned int cflag;	/* 1 : Compressed Data, 0 : Uncompressed */
		unsigned int len2;	/* inflated data size */
		char timestamp[20];
	};
};

void PrepareDirTree(const char *s0)
{
	size_t len = strlen(s0)+1;
	char *s, *t;
	if((s=(char *)malloc(len+1))==NULL || (t=(char *)malloc(len+1))==NULL){
		fprintf(stderr, "not enough memory.\n");
		exit(0);
	}
	strlcpy(s, s0, len+1);
	strlcpy(t, "", len+1);

	char *p = s;
	while(strstr(p, "\\")){
		char *q = strtok(p, "\\");
		p=p+strlen(p)+1;
		strlcat(t, q, len+1);
		strlcat(t, "\\", len+1);
		if(mkdir(t)){
			/* mkdir failed */
		}
	}

	free(t);
	free(s);
}

size_t Decompress(const unsigned char *src, size_t srclen, unsigned char *dst, size_t dstlen)
{
	unsigned char dict[4096];
	memset(dict, 0, sizeof(dict));
	int dici=4096-18;
	size_t srci=0, dsti=0;

	for(;;){
		if(srci>=srclen){
			/* input buffer underflow */
			return dsti;
		}
		unsigned char mode = src[srci++];

		for(int n=0;n<8;n++){
			if(mode&0x1){
				if(srci>=srclen){
					/* input buffer underflow */
					return dsti;
				}
				if(dsti>=dstlen){
					fprintf(stderr, "output buffer overflow detected.\n");
					return dsti;
				}
				dst[dsti++] = dict[dici++] = src[srci++];
				dici%=sizeof(dict);
			}else{
				if((srci+1)>=srclen){
					/* input buffer underflow */
					return dsti;
				}
				int c=(src[srci+1]&0x0f)+3;
				int refi=((src[srci+1]&0xf0)<<4) | (src[srci]&0xff);
				srci+=2;

				while(c--){
					if(dsti>=dstlen){
						fprintf(stderr, "output buffer overflow detected.\n");
						return dsti;
					}
					dst[dsti++] = dict[dici++] = dict[refi++];
					dici%=sizeof(dict);
					refi%=sizeof(dict);
				}
			}
			mode>>=1;
		}
	}
	return dsti;
}

void Extract(const char *fn, const union tHeader0 *idx0, const union tHeader1 *idx)
{
	FILE *src, *dst;
	if(((src=fopen(fn, "rb"))==NULL) || ((dst=fopen(idx->filename, "wb"))==NULL)){
		fprintf(stderr, "file open error.\n");
		exit(0);
	}
	fseek(src, (idx0->posData+idx->pos), SEEK_SET);

	unsigned char *srcbuf = (unsigned char *)malloc(idx->len);
	if(srcbuf==NULL){
		fprintf(stderr, "not enough memory.\n");
		exit(0);
	}
	if(fread(srcbuf, 1, idx->len, src) != idx->len){
		fprintf(stderr, "file read error.\n");
		exit(0);
	}

	if(idx->cflag){
		unsigned char *dstbuf = (unsigned char *)malloc(idx->len2);
		if(dstbuf==NULL){
			fprintf(stderr, "not enough memory.\n");
			exit(0);
		}
		if(Decompress(srcbuf, idx->len, dstbuf, idx->len2) != idx->len2){
			fprintf(stderr, "decompress failed.\n");
			exit(0);
		}
		if(fwrite(dstbuf, 1, idx->len2, dst) != idx->len2){
			fprintf(stderr, "file write error.\n");
			exit(0);
		}
		free(dstbuf);
	}else{
		if(fwrite(srcbuf, 1, idx->len, dst) != idx->len){
			fprintf(stderr, "file write error.\n");
			exit(0);
		}
	}

	free(srcbuf);
	fclose(dst);
	fclose(src);
}


int main(int argc, char *argv[])
{
	if(argv[1]==NULL){
		fprintf(stderr, "usage : %s (ac archive file)\n", argv[0]);
		exit(0);
	}

	FILE *fp;
	char *fn=argv[1];	/* (eg.)"C:\\Program Files\\LostPassage\\Data.ac1" */
	if((fp=fopen(fn, "rb"))==NULL){
		fprintf(stderr, "file open error.\n");
		exit(0);
	}

	union tHeader0 idx0;
	if(fread(idx0.img, 1, sizeof(idx0.img), fp) != sizeof(idx0.img)){
		fprintf(stderr, "file read error.\n");
		exit(0);
	}
	printf("%u Indexes\n", idx0.idx);

	for(unsigned int i=0;i<idx0.idx;i++){
		union tHeader1 idx;
		if(fread(idx.img, 1, sizeof(idx.img), fp) != sizeof(idx.img)){
			fprintf(stderr, "file read error.\n");
			exit(0);
		}
		for(int j=0;j<sizeof(idx.img);j++){
			idx.img[j] = 0xff^((((idx.img[j])>>4)&0x0f)|(((idx.img[j])<<4)&0xf0));
		}

		printf("[%u/%u]\n", i+1, idx0.idx);
		printf("FILENAME : %s\n", idx.filename);
		printf("ADDRESS  : 0x%08x\n", idx0.posData+idx.pos);
		printf("LENGTH   : 0x%08x(%u Bytes)\n", idx.len, idx.len);
		if(idx.cflag){
			printf("E.LENGTH : 0x%08x(%u Bytes)\n", idx.len2, idx.len2);
		}
		printf("TIMESTAMP: %s\n", idx.timestamp);
		printf("\n");

		PrepareDirTree(idx.filename);
		Extract(fn, &idx0, &idx);
	}

	fclose(fp);
	return 0;
}
